# from . import *
import jinja2
import confuse
from .util import *
from .safe_eval import safe_eval


def finalize_config(thing):
    return thing.get() if isinstance(thing, confuse.ConfigView) else thing

env = jinja2.Environment(finalize=finalize_config)


class Subview(confuse.Subview):
    _references = []
    def resolve(self):
        return ((self._templated_value(value), s) for value, s in super().resolve())

    def _templated_value(self, value):
        with detect_circular(self._references, value):
            # convert value
            if isinstance(value, list):
                value = [self._templated_value(v) for v in value]
            elif isinstance(value, dict):
                value = {k: self._templated_value(v) for k, v in value.items()}
            elif isinstance(value, str):
                value = safe_eval(env.from_string(value).render(C=self.root()))
            return value



SubviewBackup = confuse.Subview
def enable():
    confuse.Subview = Subview

def disable():
    confuse.Subview = SubviewBackup



'''

Jinja Result Parsing - converts str(dict) => dict

Taken from ansible:
https://github.com/ansible/ansible/blob/308723c3ca1122363e419070f1fa1d76ff5611a9/lib/ansible/template/safe_eval.py

'''

import ast
import sys
from six import string_types
from six.moves import builtins
from .util import InvalidExpression, CircularReference

# from ansible import constants as C
# from ansible.plugins.loader import filter_loader, test_loader

# define certain JSON types
# eg. JSON booleans are unknown to python eval()
OUR_GLOBALS = {
    '__builtins__': {},  # avoid global builtins as per eval docs
    'true': True, 'false': False, 'null': None,
    'True': True, 'False': False, 'None': None,
}

# this is the whitelist of AST nodes we are going to
# allow in the evaluation.
SAFE_NODES = set((
    ast.Add,
    ast.BinOp,
    ast.Compare,
    ast.Dict,
    ast.Div,
    ast.Expression,
    ast.List,
    ast.Load,
    ast.Mult,
    ast.Num,
    ast.Name,
    ast.Set,
    ast.Str,
    ast.Sub,
    ast.USub,
    ast.Tuple,
    ast.UnaryOp,
))

# And in Python 3.6 too, although not encountered until Python 3.8, see https://bugs.python.org/issue32892
if sys.version_info[:2] >= (3, 4):
    SAFE_NODES.add(ast.NameConstant)
if sys.version_info[:2] >= (3, 6):
    SAFE_NODES.add(ast.Constant)

class CleansingNodeVisitor(ast.NodeVisitor):
    def __init__(self, whitelist=None, *a, **kw):
        self.whitelist = whitelist

    def generic_visit(self, node, inside_call=False):
        if type(node) not in SAFE_NODES:
            raise InvalidExpression()
        if isinstance(node, ast.Call):
            inside_call = True
        elif self.whitelist is not None and isinstance(node, ast.Name) and inside_call:
            # Disallow calls to builtin functions that we have not vetted
            # as safe.  Other functions are excluded by setting locals in
            # the call to eval() later on iterate over all child nodes
            if hasattr(builtins, node.id) and node.id not in self.whitelist:
                raise Exception("invalid function: %s" % node.id)

        for child_node in ast.iter_child_nodes(node):
            self.generic_visit(child_node, inside_call)


def safe_eval(expr, locals=None, circular=[]):
    '''
    This is intended for allowing things like:
    with_items: a_list_variable
    Where Jinja2 would return a string but we do not want to allow it to
    call functions (outside of Jinja2, where the env is constrained).
    Based on:
    http://stackoverflow.com/questions/12523516/using-ast-and-whitelists-to-make-pythons-eval-safe
    '''
    locals = {} if locals is None else locals
    if not isinstance(expr, string_types): # already templated to a datastructure, perhaps?
        return expr

    # filter_list = []
    # for filter_ in filter_loader.all():
    #     filter_list.extend(filter_.filters().keys())
    #
    # test_list = []
    # for test in test_loader.all():
    #     test_list.extend(test.tests().keys())
    #
    # CALL_WHITELIST = C.DEFAULT_CALLABLE_WHITELIST + filter_list + test_list
    cnv = CleansingNodeVisitor() # CALL_WHITELIST
    try:
        parsed_tree = ast.parse(expr, mode='eval')
        cnv.visit(parsed_tree)
        # Note: passing our own globals and locals here constrains what
        # callables (and other identifiers) are recognized.  this is in
        # addition to the filtering of builtins done in CleansingNodeVisitor
        return eval(compile(parsed_tree, expr, 'eval'), OUR_GLOBALS, dict(locals))
    except InvalidExpression:
        raise InvalidExpression("invalid expression ({})".format(expr)) from e
    except Exception as e:
        return expr



import setuptools

USERNAME = 'beasteers'
NAME = 'confuse_jinja'

setuptools.setup(
    name=NAME,
    version='0.0.1',
    description='',
    long_description=open('README.md').read().strip(),
    long_description_content_type='text/markdown',
    author='Bea Steers',
    author_email='bea.steers@gmail.com',
    url='https://github.com/{}/{}'.format(USERNAME, NAME),
    packages=setuptools.find_packages(),
    # entry_points={'console_scripts': ['{name}={name}:main'.format(name=NAME)]},
    install_requires=['confuse'],
    license='MIT License',
    keywords='')



from contextlib import contextmanager


class InvalidExpression(Exception):
    pass

class CircularReference(Exception):
    pass


@contextmanager
def detect_circular(refs, value):
    if value in refs:
        raise CircularReference('circular reference ({}) - reference stack: {}'.format(
            value, refs))
    refs.append(value)

    yield
    assert refs.pop() == value # NOTE: this should always be true ????
